import java.io.IOException;
import java.io.InputStream;

/**
 * Multi markable buffered input stream.
 *
 * Dieser InputStream unterttzt beliebig viele Markierungen
 * im Gegensatz zum {@link java.io.BufferedInputStream}.
 *
 * Dies ist notwendig fr die Benutzung im
 * {@link QXmlParser}, um einen Blick nach vorn (look ahead)
 * zum Beispiel um festzustellen, ob jetzt ein Element ( &lt; )
 * oder ein Kommentar (&lt;!-- ) beginnt.
 *
 * Ein Problem bei der Realisierung dieser Klasse war der zu
 * wendende Puffer. Ein Puffer konstanter Grsse msste fr
 * den maximalen Abstand zwischen den Markierungen aufgerundet
 * auf die Lese-Block-Grsse  ausgelegt werden und wrde sehr
 * viel Speicher beanspruchen. Der maximale Abstand zwischen
 * erster Markierung und dem Pufferende msste bekannt sein.
 *
 * Ein dynamisches Anpassen des Puffers (Byte-Array) wrde das
 * Umkopieren aller Daten vom alten Puffer in den neu allokierten
 * Puffer erfordern, was zum Stocken des aufrufenden Programmes
 * fhren wrde und allgemein inperformant ist. Im Moment des
 * Umkopierens wrde der Speicherplatz fr beide Puffer bentigt
 * werden.
 *
 * Als Lsung dieser Probleme ist mir folgender Aufbau des Puffers
 * aus einer einfach verlinkten Liste eingefallen.
 *
 * Beim Einlesen des dekorierten {@link java.io.InputStream}
 * wird jeweils ein Byte Array in der jeweiligen Puffer-Grsse
 * angelegt.
 *
 * Mehrere Puffer sind als einfach verlinkte Liste verbunden.
 *
 * Nun bestetht das Problem, dass alte Puffer freigegeben
 * werden mssen, um den Speicher freizubekommen (garbage
 * collection).
 *
 * Der MultiMarkableBufferedInputStream hat eine Referenz auf den
 * jeweils aktuellen Puffer mit der Referenz auf den nachfolgenden
 * Puffer.
 *
 * Falls beim Lesen das Pufferende erreicht wird, legt diese Java-Klasse
 * einen NachfolgePuffer an, befllt diesen und entnimmt aus diesem die
 * zu lesenden Bytes.
 *
 * Der neue Puffer wird der aktuelle Puffer und der Vorgngerpuffer wird
 * nicht mehr referenziert. Damit kann er vom Garbage Collector entsorgt
 * werden.
 *
 * Falls das aufrufende Java-Programm eine Markierung setzen will, bekommt
 * es ein Objekt mit der Referenz auf den jeweiligen Puffer und die Position
 * im Puffer sowie auf eventuelle Nachfolge-Puffer.
 *
 * Durch den Link auf den Nachfolgepuffer werden alle nachfolgenden Puffer,
 * egal wie viele, mit referenziert und knnen nicht vom Garbage Collector
 * entsorgt werden.
 *
 * Wenn das aufrufende Java-Programm zu einer Markierung zurckkehren will,
 * ruft es die {@link #resetMark(Mark)}-Methode mit dem {@link Mark}-Objekt
 * auf. Dies setzt den bergebenen (markierten) Puffer als aktuellen Puffer
 * mitsamt aller Nachfolgepuffer und aktualisiert die Position.
 *
 * Wenn das aufrufende Java-Programm ein Mark-Objekt lscht (die Referenz) oder
 * es aus dem Scope fllt, kann der Garbage Collector den damit verbundenen
 * Puffer und alle Nachfolge-Puffer entsorgen, die nicht von anderen
 * Mark-Objekten oder in dieser Klasse referenziert werden.
 *
 * Dadurch wird erreicht,
 *
 * dass es beliebig viele Markierungen geben kann, deren Puffer und
 * eventuellen Nachfolge-Puffer solange wie notwendig referenziert werden,
 *
 * dass beliebig weit voraus gelesen werden kann (look ahead),
 *
 * dass nicht mehr bentigte Puffer dereferenziert werden und somit vom
 * Garbage Collector entsorgt werden knnen.
 *
 * <pre>
 *
 * Mark --------> Puffer Byte Array, Position
 *                 nextEntry
 *                  |
 *                  +------> Puffer Byte Array, Position
 *                            nextEntry
 *                             |
 *                             +------> Puffer Byte Array, Position
 *                                       nextEntry
 *                                        |
 * currentBuffListEntry ----------------> +------> Puffer Byte Array, Position
 *                                                  nextEntry
 *                                                   |
 *                                                   +------> Puffer Byte Array, Position
 *                                                             nextEntry
 *
 * <pre>
 *
 * @author Heiner Kcker
 * TODO MultiMarkableBufferedInputStream mit zeilen und spalten position
 */
public class MultiMarkableBufferedInputStream extends InputStream
{
  /** the decorated input stream */
  private /*final*/ InputStream decoratedInStream;

  /** Default buffer size */
  public static final int DEFAULT_BUFF_SIZE = 16 * 1024;

  /** the currently buffer size */
  final private int buffSize;

  /** the currently BuffListEntry */
  private BuffListEntry currentBuffListEntry;

  /** reference copy of currentBuffListEntry.buff for better performance */
  private byte[] buff;

  /** position within buffer */
  private int pos = 0;

  /** position in entire input stream, no position in buffer */
  private long posFromStart = 0;

  /**
   * @param pDecoratedInStream
   */
  public MultiMarkableBufferedInputStream(final InputStream pDecoratedInStream)
  {
    this( pDecoratedInStream, DEFAULT_BUFF_SIZE );
  }

  /**
   * @param pDecoratedInStream
   * @param pBuffSize
   */
  public MultiMarkableBufferedInputStream(final InputStream pDecoratedInStream,
      final int pBuffSize)
  {
    super();
    if ( pDecoratedInStream == null )
    {
        throw new NullPointerException();
    }
    this.decoratedInStream = pDecoratedInStream;
    if ( pBuffSize < 1 )
    {
        throw new IllegalArgumentException( "pBuffSize: " + pBuffSize );
    }

    this.buffSize = pBuffSize;
  }

  /**
   * @return Returns the posFromStart.
   */
  public final long getPosFromStart()
  {
    return this.posFromStart;
  }

  /**
   * @see java.io.InputStream#read()
   */
  public int read()
    throws IOException
  {
    if (currentBuffListEntry == null)
    {
      //first call
      final byte[] newBuffArr = buffFetch();
      if (newBuffArr == null)
      {
        return -1;
      }
      this.currentBuffListEntry = new BuffListEntry( newBuffArr );
      //currentBuffListEntry.buff in member kopieren fr more performance
      buff = newBuffArr;
    }
    if (buff == null)
    {
      //no data exists
      return -1;
    }
    if (pos < buff.length)
    {
      final byte by = buff[ pos ];
      pos++;
      posFromStart++;
      return by & 0xff;
    }

    //end of buffer reached

    if (currentBuffListEntry.nextEntry != null)
    {
      // es gibt einen Nachfolgepuffer
      // Weiterschalten zum Nachfolgepuffer
      currentBuffListEntry = currentBuffListEntry.nextEntry;
      //currentBuffListEntry.buff in member kopieren fr more performance
      buff = currentBuffListEntry.buff;
      this.pos = 0;
      return read();
    }

    //neuen Block vom decorierten Stream holen
    final byte[] newBuffArr = buffFetch();
    if (newBuffArr == null)
    {
      this.currentBuffListEntry = null;
      this.buff = null;
      return -1;
    }
    final BuffListEntry newBuffListEntry = new BuffListEntry( newBuffArr );
    //link the buffers
    this.currentBuffListEntry.nextEntry = newBuffListEntry;
    this.currentBuffListEntry = newBuffListEntry;
    //currentBuffListEntry.buff in member kopieren fr more performance
    this.buff = newBuffArr;
    this.pos = 0;
    return read();
  }

  /**
   * selbst implementiert fr more performance.
   *
   * @see java.io.InputStream#read(byte[], int, int)
   */
  public int read(
      final byte[] pByteArr,
      final int pOff,
      final int pLen)
    throws IOException
  {

    // return super.read(pB, pOff, pLen);
    if ((pOff < 0) || (pOff > pByteArr.length) || (pLen < 0)
        || ((pOff + pLen) > pByteArr.length) || ((pOff + pLen) < 0))
    {
      throw new IndexOutOfBoundsException();
    }
    else if (pLen == 0)
    {
      return 0;
    }

    if (currentBuffListEntry == null)
    {
      //first call
      final byte[] newBuffArr = buffFetch();
      if (newBuffArr == null)
      {
        return -1;
      }
      this.currentBuffListEntry = new BuffListEntry( newBuffArr );
      buff = newBuffArr;
    }
    if (buff == null)
    {
      //no data exists
      return -1;
    }

    int size = 0;
    while ( size < pLen )
    {
      if (pos < buff.length)
      {
        final int copySize = Math.min( pLen - size , buff.length - pos );

//System.out.println( "buff.length: "+ buff.length );
//System.out.println( "pos: "+ pos );
//System.out.println( "pByteArr.length: "+ pByteArr.length );
//System.out.println( "size: "+ size );
//System.out.println( "copySize: "+ copySize );

        System.arraycopy(buff, pos, pByteArr, size, copySize);
        pos += copySize;
        posFromStart += copySize;
        size += copySize;
      }
      else
      {
        //end of buffer reached
        if (currentBuffListEntry.nextEntry != null)
        {
          // es gibt einen Nachfolgepuffer
          // Weiterschalten zum Nachfolgepuffer
          currentBuffListEntry = currentBuffListEntry.nextEntry;
          buff = currentBuffListEntry.buff;
          this.pos = 0;
          continue; //return read();
        }

        //neuen Block vom decorierten Stream holen
        final byte[] newBuffArr = buffFetch();
        if (newBuffArr == null)
        {
          this.currentBuffListEntry = null;
          this.buff = null;
          return -1;
        }
        final BuffListEntry newBuffListEntry = new BuffListEntry( newBuffArr );
        //link the buffers
        this.currentBuffListEntry.nextEntry = newBuffListEntry;
        this.currentBuffListEntry = newBuffListEntry;
        this.buff = newBuffArr;
        this.pos = 0;
        continue; //return read();
      }
    }
    return size;
  }

  /**
   * selbst implementiert fr more performance.
   *
   * @see java.io.InputStream#skip(long)
   */
  public long skip(final long pCount) throws IOException
  {
    //return super.skip(pCount);
    long size = 0;
    while ( size < pCount )
    {
      if (pos < buff.length)
      {
        final long stepSize = Math.min( pCount - size , buff.length - pos );
        pos += stepSize;
        posFromStart += stepSize;
        size += stepSize;
      }
      else
      {
        //force buffer fetch per calling read method
        final int nextByte = read();
        if ( nextByte < 0 )
        {
          return size;
        }
        size++;
      }
    }
    return size;
  }

  /**
   * Befllen eines neuen Puffers
   * durch Lesen vom dekorierten InputStream.
   *
   * Das puffer byte array wird einfach auf die restgrsse angepasst.
   *
   * @return
   * @throws IOException
   */
  private final byte[] buffFetch()
    throws IOException
  {
//System.out.println( "buffFetch: " + posFromStart );
//if ( posFromStart > 2000)       System.exit(0);
    if ( decoratedInStream == null )
    {
      throw new IOException("input stream is closed");
    }
    final byte[] newByteArr = new byte[ buffSize ];
    final int size = decoratedInStream.read( newByteArr );
    if (size < 0)
    {
      return null;
    }
    if ( size == newByteArr.length )
    {
//System.out.println( "buffFetch: " + new String( newByteArr ) );
      return newByteArr;
    }
    //resize buffer array
    final byte[] resizeByteArr = new byte[ size ];
    System.arraycopy( newByteArr, 0 , resizeByteArr , 0 , size);
//System.out.println( "buffFetch: " + new String( resizeByteArr ) );
    return resizeByteArr;
  }

  /**
   * Get Mark position.
   *
   * @return
   * @throws IOException
   */
  public final Mark getMark()
    throws IOException
  {
    if (currentBuffListEntry == null)
    {
      //first call
      final byte[] newBuffArr = buffFetch();
      if (newBuffArr == null)
      {
        //return -1;
        //return null;
        throw new IOException( "end of file");

      }
      this.currentBuffListEntry = new BuffListEntry( newBuffArr );
      this.buff = newBuffArr;
      pos = 0;
    }
    return new Mark( currentBuffListEntry, pos, posFromStart );
  }

  /**
   * Set buffer and posizion to specified Mark.
   * @param pMark
   */
  public final void resetMark(final Mark pMark)
  {
//System.out.println( "resetMark: " + pMark );
    this.currentBuffListEntry = pMark.buffListEntry;
    this.buff = this.currentBuffListEntry.buff;
    this.pos = pMark.pos;
    this.posFromStart = pMark.posFromStart;
  }

  /* (non-Javadoc)
   * @see java.io.InputStream#available()
   */
  public int available() throws IOException
  {
    if ( buff == null )
    {
      // noch nicht initialisiert
      return decoratedInStream.available();
    }

    //return (buff.length - pos) + decoratedInStream.available();

    // verkettete Elemente beachten
    int tPos = pos;
    int buffAvail = 0;
    BuffListEntry buffListEntry = currentBuffListEntry;
    while ( buffListEntry != null )
    {
      buffAvail += buffListEntry.buff.length - tPos;
      buffListEntry = buffListEntry.nextEntry;
      tPos = 0;
    }

    return buffAvail + decoratedInStream.available();
  }

  /**
   * @see java.io.InputStream#close()
   */
  public void close() throws IOException
  {
    decoratedInStream.close();
    this.decoratedInStream = null;
    this.currentBuffListEntry = null;
    this.buff = null;
  }

  /**
   * Debug
   */
  public String toString()
  {
    try
    {
      if (this.currentBuffListEntry != null)
      {
        return
        "posFromStart: " + this.posFromStart + "\n"
        + "pos: " + this.pos + "\n"
        + new String( leftLenient( this.currentBuffListEntry.buff , 100 ) )
        + "\n======================\n"
      + new String( subLenient( this.currentBuffListEntry.buff , this.pos , this.pos + 20 ) );
      }
      return
      "posFromStart: " + this.posFromStart + "\n"
      + "currentBuffListEntry is null";
    }
    catch (Exception e)
    {
      System.err.println( e.getMessage());
      e.printStackTrace();
    }
    return "? " + this.pos;
  }

  /**
   * get the specified number of left bytes from the
   * specified byte array.
   *
   * @param pArr
   * @param pToPos
   * @return
   */
  public static byte[] leftLenient( //--
  final byte[] pArr, //--
  final int pToPos)
  {
      if ( pToPos <= pArr.length)
      {
          return left(pArr, pToPos);
      }
      return pArr;
  }

  /**
   * like java.lang.String#substring.
   *
   * @param pArr
   * @param pFromPos
   * @return
   */
  public static final byte[] left(//--
  final byte[] pArr, //--
  final int pToPos)
  {
      final int copyLength = pToPos;
      final byte[] retArr = new byte[copyLength];
      System.arraycopy(pArr, 0, retArr, 0, copyLength);
      return retArr;
  }

  /**
   * @param pBs
   * @param pI
   * @param pI2
   * @return
   */
  public static byte[] subLenient(
  final byte[] pArr, //--
  final int pFromPos, //--
  final int pToPos)
  {
      final int copyLength = Math.min( pToPos , pArr.length - 1 ) - pFromPos;
      final byte[] retArr = new byte[copyLength];
      System.arraycopy(pArr, pFromPos, retArr, 0, copyLength);
      return retArr;
  }

//  public static void main(String[] args)
//  {
//  }

  /**
   * Buffer List Entry with buffer byte array,
   * position pointer and next entry reference.
   */
  private static class BuffListEntry
    implements Cloneable
  {
    /*private*/ final byte[] buff;
    /*private*/ BuffListEntry nextEntry;
    /*private*/ //int pos = 0;

    /**
     * @param pBuff
     */
    public BuffListEntry(final byte[] pBuff)
    {
      super();
      this.buff = pBuff;
    }

  }

  /**
   * Marker class for reset.
   */
  public static class Mark
  {
    private final BuffListEntry buffListEntry;

    /** position in extra variable, damit die position nicht beim nchsten read() verndert wird */
    final int pos;

    private final long posFromStart;

    /**
     * @param pBuffListEntry
     * @param pPos
     */
    public Mark(
        final BuffListEntry pBuffListEntry,
        final int pPos,
        final long pPosFromStart)
    {
      super();
      if( pBuffListEntry == null )
      {
          throw new NullPointerException();
      }

      this.buffListEntry = pBuffListEntry;
      pos = pPos;
      posFromStart = pPosFromStart;
    }

    /**
     * @return
     */
    public final long getPosFromStart()
    {
      return posFromStart;
    }
    /**
     * Debug
     */
    public String toString()
    {
      try
      {
        return
        "posFromStart: " + this.posFromStart;

//        if (this.buffListEntry != null)
//        {
//          return
//          "posFromStart: " + this.posFromStart + "\n"
//          + "pos: " + this.buffListEntry.pos + "\n"
//          + new String( leftLenient( this.buffListEntry.buff , 100 ) )
//          + "\n======================\n"
//        + new String( subLenient( this.buffListEntry.buff , this.buffListEntry.pos , this.buffListEntry.pos + 20 ) );
//        }
//        return
//        "posFromStart: " + this.posFromStart + "\n"
//        + "currentBuffListEntry is null";
      }
      catch (Exception e)
      {
        System.err.println( e.getMessage());
        e.printStackTrace();
      }
      return "? " + this.pos;
    }

  }

}
